import ujson
import psycopg2
import psycopg2.pool
import psycopg2.extras
import eventlog as eventlog

psycopg2.extras.register_default_json(globally=True, loads=ujson.loads)
psycopg2.extras.register_default_jsonb(globally=True, loads=ujson.loads)


class DbError(RuntimeError):
    pass


class DbJsonArg(psycopg2.extras.Json):
    def dumps(self, jnode):
        """
        :type jnode: dict | list
        """
        return ujson.dumps(jnode)


class DbArgs:
    def __init__(self, db_user, db_pass, db_name, db_host='localhost', db_port=5433):
        """
        :type db_user: str
        :type db_pass: str
        :type db_name: str
        :type db_port: int
        :type db_host: str
        """
        self.db_user = db_user
        self.db_pass = db_pass
        self.db_name = db_name
        self.db_host = db_host
        self.db_port = db_port

    @staticmethod
    def deserialize(db_jnode):
        """
        :type db_jnode: dict
        :rtype: DbArgs
        :raise: KeyError
        """
        db_name = db_jnode['name']
        db_user = db_jnode['user']
        db_pass = db_jnode['pass']
        db_host = db_jnode['host']
        db_port = db_jnode['port']
        return DbArgs(db_user, db_pass, db_name, db_host, db_port)


class DbPostArgs:
    def __init__(self, post_ident, post_name, post_about):
        """
        :type post_ident: str
        :type post_name: str
        """
        self.ident = post_ident
        self.name = post_name
        self.about = post_about

    @staticmethod
    def deserialize(post_jnode):
        """
        :type post_jnode: dict
        :rtype: DbPostArgs
        """
        post_ident = post_jnode['uuid']
        post_name = post_jnode['name']
        return DbPostArgs(post_ident, post_name, post_about='')


class DbChannelArgs:
    def __init__(self, ident, name='', about=''):
        """
        :type ident: int
        :type name: str
        :type about: str
        """
        self.ident = ident
        self.name = name
        self.about = about

    @staticmethod
    def deserialzie(post_jnode):
        """
        :type post_jnode: dict
        :rtype: DbChannelArgs
        :raises: DbError
        """
        ident = post_jnode.get('Channel')
        if type(ident) not in [int]:
            eventlog.log_error('fail: param #/Channel has invalid value')
            raise DbError

        return DbChannelArgs(ident)


def with_db(db_args, handler):
    """
    :type db_args:DbArgs 
    :type handler: (Any) -> Any
    :rtype: Any
    :raises: DbError
    """
    try:
        db_link = psycopg2.connect(
            database=db_args.db_name, user=db_args.db_user, password=db_args.db_pass,
            host=db_args.db_host, port=db_args.db_port)
        try:
            return handler(db_link)
        finally:
            db_link.close()

    except psycopg2.Error as x:
        eventlog.log_debug('fail: db (%s)' % x)
        raise DbError


def edit_post(db_link, post_args):
    """
    :type db_link: Any
    :type post_args: DbPostArgs
    :raises: DbError
    """
    api_query_jnode = {
        '@header': {
            'vendor': 'VIT',
            'class': 'vitml.edge.api',
            'version': '1.0',
            'object': 'post',
            'action': 'add'
        },
        'guid': post_args.ident,
        'name': post_args.name,
        'about': post_args.about
    }

    with db_link.cursor() as db_cursor:
        try:
            db_cursor.execute("SELECT api_post_add(%s)", [DbJsonArg(api_query_jnode)])
            api_reply_jnode = db_cursor.fetchone()[0]
        except psycopg2.Error as x:
            eventlog.log_debug('fail: api_post_add (%s)' % x)
            raise DbError

        if int(api_reply_jnode['errcode']) != 0:
            eventlog.log_debug('fail: api_post_add (code=%s, text=<%s>)'
                               % (api_reply_jnode['errcode'], api_reply_jnode['errmsg']))
            raise DbError


def edit_channel(db_link, post_args, channel_args):
    """
    :type db_link: Any
    :type post_args: DbPostArgs
    :type channel_args: DbChannelArgs
    :raises: DbError
    """
    api_query_jnode = {
        '@header': {
            'vendor': 'VIT',
            'class': 'vitml.edge.api',
            'version': '1.0',
            'object': 'channel',
            'action': 'add'
        },
        'post': post_args.ident,
        'channel': channel_args.ident,
        'name': channel_args.name,
        'description': channel_args.about
    }

    with db_link.cursor() as db_cursor:
        db_cursor.execute("SELECT api_channel_add(%s)", [DbJsonArg(api_query_jnode)])
        api_reply_jnode = db_cursor.fetchone()[0]

        if int(api_reply_jnode['errcode']) != 0:
            eventlog.log_debug('fail: api_channel_add (code=%s, text=<%s>)'
                               % (api_reply_jnode['errcode'], api_reply_jnode['errmsg']))
            raise DbError
