import json
import os
import errno
from os.path import dirname as path_dirname, exists as path_exists
import shutil
import fcntl

from ujson import dump as ujson_dump
from . import eventlog


# noinspection PyBroadException
def try_lock_pid_file(file_path):
    """
    :type file_path: str
    :rtype: bool
    """
    try:
        for retry in range(2):
            try:
                # Request exclusive (EX) non-blocking (NB) advisory lock.
                fd = os.open(file_path, os.O_CREAT | os.O_WRONLY)
                fcntl.lockf(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
                os.write(fd, str(os.getpid()))
                return True

            except OSError as x:
                if x.errno != errno.ENOENT:
                    raise

            try:
                os.makedirs(os.path.dirname(file_path))
            except Exception:
                pass

    except Exception:
        pass

    return False


def try_rm_pid_file(pid_file_path):
    try:
        os.remove(pid_file_path)
        return True
    except IOError:
        pass

    return False


def silent_mkdir_r(target_path):
    """
    :type target_path: str
    """
    if not target_path:
        return

    try:
        os.makedirs(target_path)

    except (OSError, TypeError) as x:
        eventlog.log_debug("fail: make %s (%s)" % (target_path, x.message))


def silent_rm_r(path):
    if not path:
        return

    error = True
    try:
        os.remove(path)
        error = False
    except EnvironmentError:
        pass

    if error:
        try:
            shutil.rmtree(path, ignore_errors=True)
        except EnvironmentError:
            eventlog.log_debug_last_exc()


def silent_mv(src, dst):
    try:
        shutil.move(src, dst)
    except EnvironmentError as e:
        if e.errno == errno.EEXIST:
            silent_rm_r(dst)
            silent_mv(src, dst)
        elif e.errno != errno.ENOENT:  # errno.ENOENT = no such filename or directory
            raise  # re-raise exception if a different error occured


def edit_dir(target_dir, write_handler, *args, **kwargs):
    try_count = 2
    dir_present = True
    while try_count > 0 and dir_present:
        # noinspection PyBroadException
        try:
            write_handler(*args, **kwargs)
            break

        except IOError as x:
            if x.errno == errno.ENOENT:
                dir_present = False
            else:
                eventlog.log_debug('fail: write-handler (%s)' % x)

        except:
            eventlog.log_debug_last_exc()

        if not dir_present:
            silent_mkdir_r(target_dir)
            dir_present = True

        try_count -= 1

    if not try_count:
        raise RuntimeError


def write_json_file(jnode, filepath):
    """
    :type jnode: dict
    :type filepath: str
    """
    with open(filepath, 'w') as json_file:
        # Use ensure_ascii=false to output UTF-8.
        ujson_dump(jnode, json_file, ensure_ascii=False)


def write_jpeg_file(rgba_pil_image, filepath):
    """
    :type rgba_pil_image: PIL.Image.Image
    :type filepath: str
    """
    rgba_pil_image.save(filepath, 'JPEG', quality=60)


def write_rgba_file(rgba_pil_image, filepath):
    """
    :type rgba_pil_image: PIL.Image.Image
    :type filepath: str
    """
    with open(filepath, 'wb') as f:
        f.write(rgba_pil_image.tobytes())


def create_schema_symlink(schema_target_filepath, ui_jnode_filepath):
    """
    :type schema_target_filepath: str
    :type ui_jnode_filepath: str
    """
    try:
        with open(ui_jnode_filepath, 'r') as json_file:
            ui_jnode = json.load(json_file)
        schema_source_filepath = ui_jnode['ro']['schema-abspath']
    except Exception as x:
        eventlog.log_debug("fail: load ui-jnode (%s)" % x.message)
        raise

    try:
        os.remove(schema_target_filepath)
    except OSError as x:
        if x.errno != errno.ENOENT:
            eventlog.log_debug('fail: rm %s' % x.message)
            raise

    try:
        eventlog.log_debug('try: ln -s %s %s' % (schema_source_filepath, schema_target_filepath))
        silent_mkdir_r(path_dirname(schema_target_filepath))
        os.symlink(schema_source_filepath, schema_target_filepath)
        eventlog.log_debug('done')
    except OSError as x:
        eventlog.log_debug('fail: ln -s %s %s (%s)' % (schema_source_filepath, schema_target_filepath, x.message))
        raise


def create_schema_symlink2(schema_target_filepath, schema_source_filepath):
    """
    :type schema_target_filepath: str
    :type schema_source_filepath: str
    """
    try:
        os.remove(schema_target_filepath)
    except OSError as x:
        if x.errno != errno.ENOENT:
            eventlog.log_debug('fail: rm %s' % x.message)
            raise

    try:
        eventlog.log_debug('try: ln -s %s %s' % (schema_source_filepath, schema_target_filepath))
        silent_mkdir_r(path_dirname(schema_target_filepath))
        os.symlink(schema_source_filepath, schema_target_filepath)
        eventlog.log_debug('done')
    except OSError as x:
        eventlog.log_debug('fail: ln -s %s %s (%s)' % (schema_source_filepath, schema_target_filepath, x.message))
        raise


def create_jnode_from_default(def_jnode_filepath, jnode_filepath):
    """
    :type def_jnode_filepath: str
    :type jnode_filepath: str
    """
    if not def_jnode_filepath or not jnode_filepath:
        eventlog.log_debug('fail: empty jnode filepath')
        raise RuntimeError

    if not path_exists(jnode_filepath):
        # noinspection PyBroadException
        try:
            eventlog.log_debug('try: cp ui-jnode from %s to %s' % (def_jnode_filepath, jnode_filepath))
            with open(def_jnode_filepath, 'r') as json_file:
                def_ui_jnode = json.load(json_file)

            silent_mkdir_r(path_dirname(jnode_filepath))
            with open(jnode_filepath, 'w') as json_file:
                json.dump(def_ui_jnode, json_file)
            eventlog.log_debug('done')

        except:
            eventlog.log_debug('fail: create ui-jnode from %s to %s' % (def_jnode_filepath, jnode_filepath))
            raise
