from os import listdir, link, unlink
from os.path import join
from types import ImageRef

import fs_helpers

import eventlog as eventlog
import edge_rawutil.rawutil as rawutil


class ImageFileQueue:
    def __init__(self, twisted_reactor, target_codec, root_dir, min_file_count, max_file_count, log_level):
        """
        :type target_codec: str | None
        :type root_dir: str
        :type min_file_count: int
        :type max_file_count: int
        :type log_level: str
        """
        self._root_dir = root_dir
        self._min_file_count = min_file_count
        self._max_file_count = max_file_count

        self._target_codec = target_codec

        self._rawutil = rawutil.RawUtil.create(rawutil.MAX_WIDTH, rawutil.MAX_HEIGHT, log_level)
        twisted_reactor.addSystemEventTrigger('before', 'shutdown', self._rawutil.destroy)

        fs_helpers.silent_mkdir_r(self._root_dir)
        file_name_list = map(int, listdir(self._root_dir))
        file_name_list.sort(reverse=False)

        self._last_file_suffix = file_name_list.pop() if file_name_list else 0
        self._expected_file_count = len(file_name_list)

    @staticmethod
    def _cleanup(root_dir, min_file_count):
        """
        :type root_dir: str
        :type min_file_count: int
        :rtype: int
        """
        file_name_list = map(int, listdir(root_dir))
        file_name_list.sort(reverse=False)

        actual_file_count = len(file_name_list)
        for file_name in file_name_list:
            if actual_file_count <= min_file_count:
                break

            fs_helpers.silent_rm_r(join(root_dir, str(file_name)))
            actual_file_count -= 1

        return actual_file_count

    def _get_next_filepath(self):
        """
        :rtype: str
        """
        self._last_file_suffix += 1
        next_filename = join(self._root_dir, str(self._last_file_suffix))
        return next_filename

    def add_image_ref(self, image_ref):
        """
        :type image_ref: ImageRef
        :rtype: ImageRef
        """
        next_filepath = self._get_next_filepath()

        if self._target_codec is None:
            fs_helpers.edit_dir(self._root_dir, link, image_ref.url, next_filepath)

        elif image_ref.codec == 'JPEG' and self._target_codec == 'JPEG':
            fs_helpers.edit_dir(self._root_dir, link, image_ref.url, next_filepath)

        else:
            rgba_pil_image = self._rawutil.handle_image(
                image_ref.url, image_ref.codec, image_ref.width, image_ref.height)
            if not rgba_pil_image:
                raise RuntimeError('convert raw-image to rgba')

            if self._target_codec == 'JPEG':
                fs_helpers.edit_dir(self._root_dir, fs_helpers.write_jpeg_file, rgba_pil_image, next_filepath)

            elif self._target_codec == 'RGBA':
                fs_helpers.edit_dir(self._root_dir, fs_helpers.write_rgba_file, rgba_pil_image, next_filepath)

        self._expected_file_count += 1
        if self._expected_file_count >= self._max_file_count:
            self._expected_file_count = self._cleanup(self._root_dir, self._min_file_count)

        return ImageRef(ts=image_ref.ts, url=next_filepath, codec=self._target_codec)

    def add_image(self, ts, rgba_pil_image):
        """
        :type ts: int | None
        :type rgba_pil_image: PIL.Image.Image
        :rtype: ImageRef
        """
        next_filepath = self._get_next_filepath()

        if self._target_codec == 'JPEG':
            fs_helpers.edit_dir(self._root_dir, fs_helpers.write_jpeg_file, rgba_pil_image, next_filepath)

        elif self._target_codec == 'RGBA':
            fs_helpers.edit_dir(self._root_dir, fs_helpers.write_rgba_file, rgba_pil_image, next_filepath)

        self._expected_file_count += 1
        if self._expected_file_count >= self._max_file_count:
            self._expected_file_count = self._cleanup(self._root_dir, self._min_file_count)

        return ImageRef(ts=ts, url=next_filepath, codec=self._target_codec)

    def add_blob_image(self, blob_bytes, blob_ts=None, blob_codec=None):
        """
        :type blob_bytes: str
        :type blob_ts: int | None
        :type blob_codec: str | None
        :rtype: ImageRef
        """
        next_filepath = self._get_next_filepath()

        # write file to the queue
        def write_bin_file(content_bytes):
            with open(next_filepath, 'wb') as bin_file:
                bin_file.write(content_bytes)

        fs_helpers.edit_dir(self._root_dir, write_bin_file, blob_bytes)

        # cleanup the queue
        self._expected_file_count += 1
        if self._expected_file_count >= self._max_file_count:
            self._expected_file_count = self._cleanup(self._root_dir, self._min_file_count)

        return ImageRef(ts=blob_ts, url=next_filepath, codec=blob_codec)

    def remove_image(self, image_ref):
        """
        :type image_ref: LrImageRef
        """
        # assert image_ref.codec == 'JPEG' #todo : move to where called

        # noinspection PyBroadException
        try:
            unlink(image_ref.url)
            self._expected_file_count -= 1
        except:
            eventlog.log_debug_last_exc()


class JsonFileQueue:
    def __init__(self, args):
        """
        :type args: QueueArgs
        """
        self._root_dir = args.endpoint
        self._min_file_count = args.min_file_count
        self._max_file_count = args.max_file_count

        fs_helpers.silent_mkdir_r(self._root_dir)
        file_name_list = map(int, listdir(self._root_dir))
        file_name_list.sort(reverse=True)

        self._last_file_suffix = file_name_list.pop() if file_name_list else 0
        self._expected_file_count = len(file_name_list)

    @staticmethod
    def _cleanup(root_dir, min_file_count):
        """
        :type root_dir: str
        :type min_file_count: int
        :rtype: int
        """
        file_name_list = map(int, listdir(root_dir))
        file_name_list.sort(reverse=False)

        actual_file_count = len(file_name_list)
        for file_name in file_name_list:
            if actual_file_count <= min_file_count:
                break

            fs_helpers.silent_rm_r(join(root_dir, str(file_name)))
            actual_file_count -= 1

        return actual_file_count

    def _get_next_filepath(self):
        """
        :rtype: str
        """
        self._last_file_suffix += 1
        next_filename = join(self._root_dir, str(self._last_file_suffix))
        return next_filename

    def add_jnode(self, jnode):
        """
        :type jnode: dict
        """
        next_filepath = self._get_next_filepath()
        fs_helpers.edit_dir(self._root_dir, fs_helpers.write_json_file, jnode, next_filepath)

        self._expected_file_count += 1
        if self._expected_file_count >= self._max_file_count:
            self._expected_file_count = self._cleanup(self._root_dir, self._min_file_count)
